use std::fs;
use std::io::{stdin, ErrorKind, Read, Write};
use std::os::unix::net::{UnixListener, UnixStream};
use std::path::PathBuf;
use std::thread;

use clap::{Parser, Subcommand};
use color_eyre::eyre::Context;
use color_eyre::Result;
use serde::Deserialize;

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
struct Cli {
    /// Set the level of verbosity
    #[arg(short, long, action = clap::ArgAction::Count)]
    verbose: u8,
    /// The command to run
    #[command(subcommand)]
    command: Command,
}

#[derive(Subcommand)]
enum Command {
    /// Send a request payload to a running daemon and wait for the response
    Request,
    /// Start a daemon
    Daemon {
        /// The Kakoune session this daemon belongs to
        session: String,
    },
}

#[derive(Clone, Debug, Deserialize)]
struct Request {
    session: String,
}

fn main() -> Result<()> {
    let cli = Cli::parse();

    let runtime_dir = std::env::var("XDG_RUNTIME_DIR")
        .map(PathBuf::from)
        .unwrap_or_else(|_| "/var".into())
        .join("kak-highlight");

    match cli.command {
        Command::Daemon { session } => {
            if let Err(err) = fs::metadata(&runtime_dir) {
                if err.kind() == ErrorKind::NotFound {
                    fs::create_dir_all(&runtime_dir).wrap_err_with(|| {
                        format!(
                            "Failed to create runtime directory {}",
                            runtime_dir.display()
                        )
                    })?;
                }
            }

            let path = runtime_dir.join(session).with_extension("s");
            let listener = UnixListener::bind(&path)
                .wrap_err_with(|| format!("Failed to bind listener to {}", path.display()))?;

            for stream in listener.incoming() {
                let stream = stream.wrap_err("Failed to accept incoming connection")?;
                thread::spawn(|| {
                    if let Err(err) = handle_connection(stream) {
                        eprintln!("{:?}", err);
                    }
                });
            }

            Ok(())
        }
        Command::Request => {
            let mut buf = String::new();
            stdin()
                .read_to_string(&mut buf)
                .wrap_err("Failed to read stdin")?;
            let mut socket = {
                let req: Request = toml::from_str(&buf).wrap_err("Failed to parse request")?;
                let path = runtime_dir.join(req.session).with_extension("s");

                UnixStream::connect(&path).wrap_err_with(|| {
                    format!("Failed to connect to daemon socket {}", path.display())
                })?
            };

            socket
                .write_all(buf.as_bytes())
                .wrap_err("Failed to send response")?;

            Ok(())
        }
    }
}

fn handle_connection(mut read: impl Read) -> Result<()> {
    let mut buf = String::new();
    read.read_to_string(&mut buf)
        .wrap_err("Failed to read stdin")?;
    let req: Request = toml::from_str(&buf).wrap_err("Failed to parse request")?;

    dbg!(req);

    todo!();
}
